#pragma once

#include <assert.h>

#ifdef WIN32
#undef min
#undef max
#endif
#include "captury/compilers.h"

// some type declarations
#include <stdint.h>
#include <stddef.h>		// for size_t

typedef uint32_t uint;

typedef int64_t timestamp_t;
typedef int64_t timestamp_diff_t;

#include <limits>
#define TIMESTAMP_DUMMY_VALUE std::numeric_limits<int64_t>::min()
#define TIMESTAMP_MIN std::numeric_limits<int64_t>::min()
#define TIMESTAMP_MAX std::numeric_limits<int64_t>::max()

enum subsample_t {HIGHEST = -1, FULL = 0, HALF = 1, QUARTER = 2, EIGHTH = 3, SIXTEENTH = 4, INEXISTENT = 5, MAX = 5};

enum failureRecoveryMode_t { FAILURE_RECOVERY_OFF = 0, FAILURE_RECOVERY_CONSERVATIVE = 1, FAILURE_RECOVERY_NORMAL = 2, FAILURE_RECOVERY_PROACTIVE = 3, FAILURE_RECOVERY_DNN = 4, FAILURE_RECOVERY_ALWAYS = 5, FAILURE_RECOVERY_SNAP_POSE = 6 };

enum snapMode_t { SnapMode_BackgroundSubtractionLocal, SnapMode_BackgroundSubtractionGlobal, SnapMode_BodyPartDetectorJoints };

enum similarityDisplayType_t { similarityDisplay_Blobs, similarityDisplay_Proxies, similarityDisplay_BlobsDNN, similarityDisplay_ProxiesDNN };

enum axis_t { PLUS_X=1, MINUS_X=-1, PLUS_Y=2, MINUS_Y=-2, PLUS_Z=3, MINUS_Z=-3 };

// prefix (++orig)
inline subsample_t& operator++(subsample_t& orig)
{
	orig = static_cast<subsample_t>(orig + 1); // static_cast required because enum + int -> int
	return orig;
}

// postfix (orig++)
inline subsample_t operator++(const subsample_t& orig, int)
{
	subsample_t ret = orig;
	++ret;
	return ret;
}

class videoframe_t
{
public:
	enum validState { invalid };

	// constructor
	videoframe_t(validState vs): camera(0xFFFFFFFF), frame(0xFFFFFFFF) {}
	videoframe_t(): camera(0xFFFFFFFF), frame(0xFFFFFFFF) {}
	explicit videoframe_t(const uint c, int f) : camera(c), frame(f) {}
	// copy constructor
	videoframe_t(const videoframe_t& f)			{ camera = f.camera; frame = f.frame; }
	// operators
	videoframe_t& operator=(const videoframe_t& f)		{ camera = f.camera; frame = f.frame; return *this; }
	videoframe_t& operator+=(int delta)			{ frame += delta; return *this; }
	videoframe_t& operator-=(int delta)			{ frame -= delta; return *this; }
	videoframe_t& operator++()				{ frame++; return *this; }
	videoframe_t& operator--()				{ frame--; return *this; }
	// non-destructive operators
	videoframe_t operator+(int delta) const			{ return videoframe_t(getCamera(), frame + delta); }
	videoframe_t operator-(int delta) const			{ return videoframe_t(getCamera(), frame - delta); }
	int operator-(const videoframe_t& f) const		{ assert(getCamera() == f.getCamera()); return (getFrame() - f.getFrame()); }
	double operator*(double scale) const			{ return frame * scale; }
	double operator/(double scale) const			{ return frame / scale; }
	int operator%(int mod) const				{ return frame % mod; }
	// comparison operators
	bool operator<(const videoframe_t& f) const		{ return (camera < f.camera) || (camera == f.camera && frame < f.frame); }
	bool operator==(const videoframe_t& f) const		{ return (camera == f.camera) && (frame == f.frame); }
	bool operator<=(const videoframe_t& f) const		{ return (*this < f) || (*this == f); }
	bool operator>(const videoframe_t& f) const		{ return (camera > f.camera) || (camera == f.camera && frame > f.frame); }
	bool operator>=(const videoframe_t& f) const		{ return (*this > f) || (*this == f); }
	bool operator!=(const videoframe_t& f) const		{ return (camera != f.camera) || (frame != f.frame); }
	int getFrame() const					{ return frame; }
	void setFrame(uint f)					{ frame = f; }
	uint getCamera() const					{ return camera; }
	void setCamera(uint cam)				{ camera = cam; }
	videoframe_t& operator=(validState vs)			{ camera = 0xFFFFFFFF; frame = 0xFFFFFFFF; return *this; }
	bool operator==(validState vs) const			{ return camera == 0xFFFFFFFF && uint(frame) == 0xFFFFFFFF; }
	bool operator!=(validState vs) const			{ return !(*this == vs); }

	uint	camera;
	int	frame;
};

class trackingframe_t
{
public:
	enum validState { invalid, max };

	// constructor
	trackingframe_t(validState vs)				{ if (vs == invalid) frame = 0xFFFFFFFF; else frame = 0x7FFFFFFF; }
	trackingframe_t()					{ frame = 0xFFFFFFFF; }
	explicit trackingframe_t(const int& f) : frame(f)	{}
	// copy constructor
	trackingframe_t(const trackingframe_t& f)		{ frame = f.frame; }
	// operators
	trackingframe_t& operator=(const trackingframe_t& f)	{ frame = f.frame; return *this; }
	trackingframe_t& operator+=(int delta)			{ frame += delta; return *this; }
	trackingframe_t& operator-=(int delta)			{ frame -= delta; return *this; }
	trackingframe_t& operator++()				{ ++frame; return *this; }
	trackingframe_t& operator--()				{ --frame; return *this; }
	// non-destructive operators
	trackingframe_t operator+(int delta) const		{ return trackingframe_t(frame + delta); }
	trackingframe_t operator+(const trackingframe_t& f) const{ return trackingframe_t(frame + f.frame); }
	trackingframe_t operator-(int delta) const		{ return trackingframe_t(frame - delta); }
	int operator-(const trackingframe_t& f) const		{ return static_cast<int>(frame) - static_cast<int>(f.frame); }
	double operator*(double scale) const			{ return frame * scale; }
	double operator/(double scale) const			{ return frame / scale; }
	int operator%(int mod) const				{ return frame % mod; }
	// comparison operators
	bool operator<(const trackingframe_t& f) const		{ return frame < f.frame; }
	bool operator<=(const trackingframe_t& f) const		{ return frame <= f.frame; }
	bool operator==(const trackingframe_t& f) const		{ return frame == f.frame; }
	bool operator!=(const trackingframe_t& f) const		{ return frame != f.frame; }
	bool operator>=(const trackingframe_t& f) const		{ return frame >= f.frame; }
	bool operator>(const trackingframe_t& f) const		{ return frame > f.frame; }
	int toInt() const					{ return frame; }

	trackingframe_t& operator=(validState vs)		{ if (vs == max) frame = 0x7FFFFFFF; else frame = 0xFFFFFFFF; return *this; }
	bool operator==(validState vs) const			{ return (vs == max && uint(frame) == 0x7FFFFFFF) || (vs == invalid && uint(frame) == 0xFFFFFFFF); }
	bool operator!=(validState vs) const			{ return !(*this == vs); }

	int frame;
};

class fraction_t
{
public:
	// constructor
	fraction_t()						{}
	explicit fraction_t(double x);
	explicit fraction_t(int integer) : num(integer), denom(1)	{}
	fraction_t(int64_t numerator, int64_t denominator)	{ if (denominator < 0) simplify(-numerator, -denominator); else simplify(numerator, denominator); }
	// copy constructor
	fraction_t(const fraction_t& frac) : num(frac.num), denom(frac.denom) {}
	// operators
	fraction_t& operator=(const fraction_t& f)		{ num = f.num; denom = f.denom; return *this; }
	fraction_t& operator*=(const fraction_t& f)
	{
		if (mul_overflow(num, f.num, &num) ||
		    mul_overflow(denom, f.denom, &denom)) {
			double x = (double)num * f.num / ((double)denom * f.denom);
			*this = fraction_t(x);
		} else
			simplify(num, denom);
		return *this;
	}
	fraction_t& operator*=(const int64_t& integer)
	{
		if (mul_overflow(num, integer, &num)) {
			double x = (double)num * integer / (double)denom;
			*this = fraction_t(x);
		} else
			simplify(num, denom);
		return *this;
	}
	fraction_t& operator/=(const fraction_t& f)
	{
		if (mul_overflow(num, f.denom, &num) ||
		    mul_overflow(denom, f.num, &denom)) {
			double x = (double)num * f.denom / ((double)denom * f.num);
			*this = fraction_t(x);
		} else
			simplify(num, denom);
		return *this;
	}
	fraction_t& operator/=(const int64_t& integer)
	{
		if (mul_overflow(denom, integer, &denom)) {
			double x = (double)num / ((double)denom * integer);
			*this = fraction_t(x);
		} else
			simplify(num, denom);
		return *this;
	}
	fraction_t& operator+=(const fraction_t& f)
	{
		int64_t n, d;
		if (mul_overflow(num, f.denom, &n) ||
		    mul_overflow(denom, f.num, &d) ||
		    add_overflow(n, d, &n) ||
		    mul_overflow(denom, f.denom, &d))
			*this = fraction_t(((double)num * f.denom + (double)f.num * denom) / ((double)denom * f.denom));
		else
			simplify(n, d);
		return *this;
	}
	fraction_t& operator-=(const fraction_t& f)
	{
		int64_t n, d;
		if (mul_overflow(num, f.denom, &n) ||
		    mul_overflow(denom, f.num, &d) ||
		    sub_overflow(n, d, &n) ||
		    mul_overflow(denom, f.denom, &d))
			*this = fraction_t(((double)num * f.denom - (double)f.num * denom) / ((double)denom * f.denom));
		else
			simplify(n, d);
		return *this;
	}
	fraction_t& operator+=(int64_t integer)
	{
		int64_t x;
		if (mul_overflow(integer, denom, &x) ||
		    add_overflow(num, x, &x))
			*this = fraction_t(((double)num + (double)integer * denom) / denom);
		else
			simplify(x, denom);
		return *this;
	}
	fraction_t& operator-=(int64_t integer)
	{
		int64_t x;
		if (mul_overflow(integer, denom, &x) ||
		    sub_overflow(num, x, &x))
			*this = fraction_t(((double)num - (double)integer * denom) / denom);
		else
			simplify(x, denom);
		return *this;
	}
	// non-destructive operators
	fraction_t operator*(const fraction_t& f) const
	{
		fraction_t result;
		if (mul_overflow(num, f.num, &result.num) ||
		    mul_overflow(denom, f.denom, &result.denom)) {
			double x = (double)num * f.num / ((double)denom * f.denom);
			return fraction_t(x);
		} else {
			result.simplify(result.num, result.denom);
			return result;
		}
	}
	fraction_t operator/(const fraction_t& f) const
	{
		fraction_t result;
		if (mul_overflow(num, f.denom, &result.num) ||
		    mul_overflow(denom, f.num, &result.denom)) {
			double x = (double)num * f.denom / ((double)denom * f.num);
			return fraction_t(x);
		} else {
			result.simplify(result.num, result.denom);
			return result;
		}
	}
	fraction_t operator*(const int64_t& integer) const
	{
		fraction_t result(1, denom);
		if (mul_overflow(num, integer, &result.num)) {
			double x = (double)num * integer / (double)denom;
			return fraction_t(x);
		} else {
			result.simplify(result.num, result.denom);
			return result;
		}
	}
	fraction_t operator/(const int64_t& integer) const
	{
		fraction_t result(num, 1);
		if (mul_overflow(denom, integer, &result.denom)) {
			double x = (double)num / ((double)denom * integer);
			return fraction_t(x);
		} else {
			result.simplify(result.num, result.denom);
			return result;
		}
	}
	fraction_t operator+(const fraction_t& f) const
	{
		fraction_t result;
		if (mul_overflow(num, f.denom, &result.num) ||
		    mul_overflow(denom, f.num, &result.denom) ||
		    add_overflow(result.num, result.denom, &result.num) ||
		    mul_overflow(denom, f.denom, &result.denom))
			return fraction_t(((double)num * f.denom + (double)f.num * denom) / ((double)denom * f.denom));
		else
			result.simplify(result.num, result.denom);
		return result;
	}
	fraction_t operator-(const fraction_t& f) const
	{
		fraction_t result;
		if (mul_overflow(num, f.denom, &result.num) ||
		    mul_overflow(denom, f.num, &result.denom) ||
		    sub_overflow(result.num, result.denom, &result.num) ||
		    mul_overflow(denom, f.denom, &result.denom))
			return fraction_t(((double)num * f.denom - (double)f.num * denom) / ((double)denom * f.denom));
		else
			result.simplify(result.num, result.denom);
		return result;
	}
	fraction_t operator+(int64_t integer) const
	{
		fraction_t result;
		if (mul_overflow(integer, denom, &result.num) ||
		    add_overflow(num, result.num, &result.num))
			return fraction_t(((double)num + (double)integer * denom) / denom);
		else
			result.simplify(result.num, denom);
		return result;
	}
	fraction_t operator-(int64_t integer) const
	{
		fraction_t result;
		if (mul_overflow(integer, denom, &result.num) ||
		    sub_overflow(num, result.num, &result.num))
			return fraction_t(((double)num + (double)integer * denom) / denom);
		else
			result.simplify(result.num, denom);
		return result;
	}
	// comparison operators
	bool operator<(const fraction_t& f) const		{ return num*f.denom < f.num*denom; }
	bool operator<=(const fraction_t& f) const		{ return num*f.denom <= f.num*denom; }
	bool operator==(const fraction_t& f) const		{ return num*f.denom == f.num*denom; }
	bool operator!=(const fraction_t& f) const		{ return num*f.denom != f.num*denom; }
	bool operator>=(const fraction_t& f) const		{ return num*f.denom >= f.num*denom; }
	bool operator>(const fraction_t& f) const		{ return num*f.denom > f.num*denom; }
	bool operator<(const int& x) const			{ return num < x*denom; }
	bool operator<=(const int& x) const			{ return num <= x*denom; }
	bool operator==(const int& x) const			{ return num == x*denom; }
	bool operator!=(const int& x) const			{ return num != x*denom; }
	bool operator>(const int& x) const			{ return num > x*denom; }
	bool operator>=(const int& x) const			{ return num >= x*denom; }
	// accessors
	fraction_t inv() const					{ return fraction_t(denom, num); }
	double toDouble() const					{ return ((double) num) / denom; }
	int64_t toInt() const					{ if (denom == 0) return 0; return num / denom; } // round to 0
	int64_t round() const					{ if (denom == 0) return 0; return (num>0) ? (num + (((denom < 0) ? -denom : denom) >> 1)) / denom : (num - (((denom < 0) ? -denom : denom) >> 1)) / denom; }
	fraction_t abs() const					{ return fraction_t((num < 0) ? -num : num, (denom < 0) ? -denom : denom); }
	fraction_t getFractionalPart() const			{ if ((num > 0 && num < denom) || (num < 0 && -num < denom)) return *this; return fraction_t(num % denom, denom); }
	int64_t getNumerator() const				{ return num; }
	uint64_t getDenominator() const				{ return denom; }

private:
	void simplify(int64_t numerator, uint64_t denominator);

	int64_t	num;	// numerator
	int64_t	denom;	// denominator
};


typedef uint64_t shotid_t;

