#ifndef WIN32
#include <vector>
#include <string>
#include <stdio.h>
#include "captury/plugin.h"


#include <ros/ros.h>
#include <tf2/LinearMath/Quaternion.h>
#include <tf2_ros/transform_broadcaster.h>
#include <geometry_msgs/TransformStamped.h>

// NOTE: for this plugin to be loaded make sure that the ros environment has been loaded
// NOTE: this can be achieved for example with "source /opt/ros/lunar/setup.bash"

template <typename T>
static inline T sqr(T x)        { return x * x; }

// actor id -> joint names
static std::map<int, std::vector<std::string>> jointNames;
static std::map<int, std::vector<int>> parents;

struct M4x4 {
	float coeffs[16];
};

static M4x4 globalTrafo; // ROS has y=up and x=forward

static pluginInfo* plugin;

static char nameSpace[100] = "";

static void getActor(uint64_t actorId)
{
	CapturyActor* actor = liveGetActor(actorId, 0);
	std::vector<std::string>& jn = jointNames[actorId];
	jn.resize(actor->numJoints);
	std::vector<int>& ps = parents[actorId];
	ps.resize(actor->numJoints);
	for (int i = 0; i < actor->numJoints; ++i) {
		jn[i] = std::string("/") + actor->joints[i].name;
		ps[i] = actor->joints[i].parent;
	}
	for (int i = actor->numJoints - 1; i >= 0; --i) {
		if (ps[i] < 0)
			continue;

		if (jn[ps[i]] == jn[i]) { // ignore split complex joints
			int tmp = ps[ps[i]];
			ps[ps[i]] = -2;
			ps[i] = tmp;
		}
	}
}

static void actorModeChanged(uint64_t actorId, LiveActorMode mode)
{
	if (jointNames.count(actorId) != 0) // we already know this guy
		return;

	if (mode == liveSkeletonScaling || mode == liveSkeletonTracking || mode == liveSkeletonPlayback)
		getActor(actorId);
}

static void newPose(const CapturyPose* pose, int trackingQuality)
{
	static tf2_ros::TransformBroadcaster br;
	geometry_msgs::TransformStamped transformStamped;

	uint64_t deltaTimeInUsec = liveGetTime() - pose->timestamp;
	transformStamped.header.stamp = ros::Time::now();// - ros::Duration(deltaTimeInUsec * 1e-6);
	if (deltaTimeInUsec < 500000) { // less than half a second - otherwise probably playback mode
		// liveLogInfo("delta time %g %g", deltaTimeInUsec * 1e-6, pose->timestamp * 1e-6);
		transformStamped.header.stamp -= ros::Duration(deltaTimeInUsec * 1e-6);
	}

	std::vector<M4x4> invMats(pose->numTransforms); // for computing relative frames

	if (jointNames.count(pose->actor) == 0)
		getActor(pose->actor);

	for (int i = 0; i < pose->numTransforms; ++i) {
		float m[16];
		float relMat[16];
		liveTransformToMatrix(&pose->transforms[i], relMat);
		liveMatrixMatrix(globalTrafo.coeffs, relMat, m); // rotate to match ROS coordinate system convention (up=z, x=forward)

		liveInvertMatrix(m, invMats[i].coeffs);

		transformStamped.child_frame_id = jointNames[pose->actor][i];

		// compute relative frame
		int parent = parents[pose->actor][i];
		if (parent == -1) {
			transformStamped.header.frame_id = "map";
			memcpy(relMat, m, sizeof(m));
		} else if (parent < -1) {
			continue;
		} else {
			if (jointNames[pose->actor][parent] == transformStamped.child_frame_id)
				continue;
			transformStamped.header.frame_id = jointNames[pose->actor][parent];
			liveMatrixMatrix(invMats[parent].coeffs, m, relMat);
		}

		// fill message
		transformStamped.transform.translation.x = relMat[3] * 0.001f; // ROS works in meters
		transformStamped.transform.translation.y = relMat[7] * 0.001f;
		transformStamped.transform.translation.z = relMat[11] * 0.001f;

		float quat[4];
		liveMatrixToQuaternion(relMat, quat);
		float norm = std::sqrt(sqr(quat[0]) + sqr(quat[1]) + sqr(quat[2]) + sqr(quat[3]));
		transformStamped.transform.rotation.w = quat[0] / norm;
		transformStamped.transform.rotation.x = quat[1] / norm;
		transformStamped.transform.rotation.y = quat[2] / norm;
		transformStamped.transform.rotation.z = quat[3] / norm;

		br.sendTransform(transformStamped);
	}
}

static void activate(pluginInfo* info)
{
	liveLogInfo("activating ROS publisher");
	liveRegisterPoseCallback(newPose);
}

static void deactivate()
{
	liveUnregisterPoseCallback(newPose);
	liveLogInfo("deactivating ROS publisher");
}

static void deinit(pluginInfo* info)
{
	liveLogInfo("deiniting ROS publisher");
}

static void knobChanged(const char* name)
{
	if (strcmp(name, "masterUri") == 0 || strcmp(name, "namespace") == 0) {
		char uri[100];
		livePanelGetLineEditText(plugin, name, uri, 100);
		liveStoreString(plugin, name, uri);
	}
}

extern "C" unsigned char initPlugin(pluginInfo* info)
{
	if (info->size != sizeof(pluginInfo)) {
		liveLogWarning("incompatible version of CapturyLive for ROS plugin");
		return 0;
	}

	liveLogInfo("initing ROS publisher");

	plugin = info;

	info->name = "ROS";
	info->activate = activate;
	info->deactivate = deactivate;
	info->deinit = deinit;
	info->actorModeChanged = actorModeChanged;
	info->knobChanged = knobChanged;

	liveAddPanel(info);

	livePanelAddLineEdit(info, "masterUri", "ROS master URI (restart when changed):");
	char masterUri[100];
	liveRetrieveString(info, "masterUri", masterUri, 100, "http://127.0.0.1:11311");
	if (masterUri[0] == '\0') {
		liveStoreString(info, "masterUri", "http://127.0.0.1:11311");
		strcpy(masterUri, "http://127.0.0.1:11311");
	}
	livePanelSetLineEditText(info, "masterUri", masterUri);

	livePanelAddLineEdit(info, "namespace", "namespace (restart when changed):");
	liveRetrieveString(info, "namespace", nameSpace, 100, "");
	livePanelSetLineEditText(info, "namespace", nameSpace);

	int argc = 3;
	char arg0[] = "CapturyLive";
	char arg1[120];
	sprintf(arg1, "__master:=%s", masterUri);
	char arg2[120];
	sprintf(arg2, "__ns:=%s", nameSpace);
	char *argv[3] = {arg0, arg1, arg2};
	ros::init(argc, argv, "CapturyLive_tf2_broadcaster");


	memset(globalTrafo.coeffs, 0.0f, sizeof(globalTrafo.coeffs));
	globalTrafo.coeffs[2] = -1.0f; // rotate y up, x forward
	globalTrafo.coeffs[4] = -1.0f;
	globalTrafo.coeffs[9] = 1.0f;
	globalTrafo.coeffs[15] = 1.0f;

	return 1;
}

#endif
